#    "Copyright (C) 2021  Info Libera (liberinforma[at]gmail.com)"

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#30/06/2021
#Version 00
#User selection of SLMs to be used
#Loading of SLMs
#User programming of experiment
#Validation of program

#30/07/2021
#Version 01
#Implement State Machine of SLMs
#Implement State Machine of TSC


#17/08/2021
#Version 03
#Implement execCommandExp to execute commands during an experiment
#and the events from NORMAL_OPERATION state
#Implement the class dataFlowController

#25/08/2021
#Version 04
#Implement code to run an experiment
#Method waitTime for TSC

#14/12/2021
#SisProLab - Version 00

#21/12/2021
#SisProLab - Version 01
#Include SLM_04 (syringe pump)

###############################################################
#Instrument server
###############################################################

oo::class create instrumentServer {

    variable inst_server_attributes
    
    constructor { } {

    }

    #Available instruments from the variable available_resources
    
    method setAvailableInst { available_resources } {

	#puts "Method setAvailableInst of [self]"
	
	foreach slm [dict keys $available_resources] {

	    set t [dict get $available_resources $slm type]

	    if { $t == "instrument" } {

		set n [dict get $available_resources $slm name]
		dict set inst_server_attributes inst $slm name $n

	    }

	}
    }

    method getAvailableInst {} {

	#puts "Method getAvailableInst returns [dict get $inst_server_attributes inst]"

	if { [dict exists $inst_server_attributes inst] } {
	    return [dict get $inst_server_attributes inst]
	} else {
	    return 0
	}

    }

    method setDataFlowController { dfc } {

	dict set inst_server_attributes dfc $dfc
	
    }

    method getDataFlowController {} {


	return [dict get $inst_server_attributes dfc]

    }
    
    method createInstModule { slm_id } {

	#puts "Method createInstModule of [self] for slm:$slm_id"

	set obj_active_slm [instrumentModule new $slm_id]

	set state_obj_active_slm [[self] createInstState $obj_active_slm]

	$obj_active_slm setState $state_obj_active_slm

	#Agregate de instance of dataFlowController to the instrumentModule
	set data_flow_controller [[self] getDataFlowController]

	$obj_active_slm setDataFlowController $data_flow_controller
	
	dict lappend inst_server_attributes obj_active_slm $obj_active_slm

	return $obj_active_slm
	       
    }

    #This method receive the name of the object of module (::oo::Objnn)
    #and returns the corresponding id (slm_nn)
    method getIDModule { slm_obj } {

	#puts "[self] - method getIDModule"
	
	return [$slm_obj getID]

    }

    #This method receive the id of slm (slm_nn) and returns
    #the corresponding name of the object of module (::oo::Objnn)
    method getObjModule { slm_id } {

	#puts "[self] - method getObjModule for $slm_id\n"
	
	foreach obj_slm [[self] getSelectedObjSLM] {
	    
	    if { [[self] getIDModule $obj_slm] == $slm_id } {
		
		return $obj_slm
		
	    }
	    
	}

       return 0

    }
    
    method createInstState { id_obj } {

	#puts "Method createInstState of [self] for the SLM id_obj -> $id_obj\n"
	
	set obj_state [stateUNCONNECTED new $id_obj]

	return $obj_state
	
    }

    method getSelectedObjSLM {} {

	#puts "Method getSelectedObjSLM of [self]"

	if { [ dict exists $inst_server_attributes obj_active_slm ] } {
	    return [ dict get $inst_server_attributes obj_active_slm ]
	} else {
	    return 0
	}
    }

    method setAvailableCommands { available_resources } {

	#puts "\nMethod setAvailableCommands of instrumentServer ([self])\n"

	foreach active_obj_slm [[self] getSelectedObjSLM] {

	    #Check if there is at least one active slm
	    
	    if { $active_obj_slm != 0 } {
		
		#puts "SLM of $active_obj_slm -> [$active_obj_slm getID]"
		
		set slm_active [$active_obj_slm getID]
		
		foreach slm [dict keys $available_resources] {
		    
		    #puts "Loop foreach with slm -> $slm and slm_active -> $slm_active\n"
		    
		    if { $slm == $slm_active } {
			
			set c [dict get $available_resources $slm command]
			
			#puts "Commands available of instrument $slm [dict keys $c]\n"
			
			dict set inst_server_attributes $slm command $c
			
		    }
		    
		}
	    
	    }

	}

    }

    #Commands defined by the variable available_resources
    
    method getAvailableCommands {} {

	#puts "Method getAvailableCommands of instrumentServer [self]\n"

	foreach active_obj_slm [[self] getSelectedObjSLM] {
	    
	    if { $active_obj_slm != 0 } {
		
		set slm [$active_obj_slm getID]
		
		#puts "$slm [dict get $inst_server_attributes $slm command]"
		
		dict set commands_inst $slm [dict get $inst_server_attributes $slm command]
	    }
	}
	
	if { [info exists commands_inst ] } {
	    return $commands_inst
	} else {
	    return ""
	}
	
    }

    method startConnection { slm_id } {

	puts "\n#Method startConnection of [self] ([info object class [self]]) for slm_id - $slm_id\n"

	set data_flow_controller [[self] getDataFlowController]

	set id_serial_port [$data_flow_controller getSerialPortID]

	puts "\nThe dataFlowController $data_flow_controller and de id_serial_port $id_serial_port\n"

	if { ($id_serial_port == "0") } {

	    $data_flow_controller openPort

	    set id_serial_port [$data_flow_controller getSerialPortID]
	    
	}
	
	
	set obj_active_slm [[self] createInstModule $slm_id]

	set return_start_connection [$obj_active_slm applyEvent eventSTART_CONNECTION]

	#puts "\nreturn_start_connection - $return_start_connection\n"

	if { $return_start_connection == "ACK" } {
	    
	    [self] loadCommands $obj_active_slm
	    
	    set return_check_connection [[self] execCommand $obj_active_slm checkConnection NO_ARG]

	    #puts "return_check_connection - $return_check_connection"

	    if { $return_check_connection == "ACK" } {

		$obj_active_slm applyEvent eventCONNECTION_ACCEPTED

	    }
	    
	    return $return_check_connection 
	    
	} else {

	    return $return_start_connection

	}
	
    }
    
    method setLoadedCommands { id_obj_slm list_commands } {

	#puts " ****Instrument Server - Method setLoadedCommands ob_slm: $id_obj_slm commands: $list_commands **** \n"
	
	dict set inst_server_attributes loaded_commands $id_obj_slm $list_commands

    }
    
    method getLoadedCommands { id_obj_slm } {

	return [dict get $inst_server_attributes loaded_commands $id_obj_slm]
    }

    method loadCommands { obj_slm } {
	
	set list_commands [$obj_slm loadSLM]

	[self] setLoadedCommands $obj_slm $list_commands
	
	#puts " ***Instrument Server - Method loadCommands of slm $obj_slm load the commands $list_commands ***\n"
	
    }

    method getArgsCommand { obj_slm command_name } {

	set list_args_command [$obj_slm getArgsCommandSLM $command_name]

	return $list_args_command
    }

    method execCommand { obj_slm cmd list_args} {

	#puts "Method execCommand of instrumentServer [self] module:$obj_slm cmd:$cmd args:$list_args"

	
	    return [$obj_slm execCommand $cmd $list_args]

	
	}

    method applyEventSLM { slm_id event list_args} {

	#puts "Method applyEventSLM of [info object class [self]] with $slm_id $event and $list_args\n"

	set obj_slm [[self] getObjModule $slm_id]
	
	set return_apply_event [ $obj_slm applyEvent $event]

	puts "Return of apply $event in module [$obj_slm getID] is $return_apply_event\n"

	if { $return_apply_event == "ACK" } {

	    set return_exec_command [$obj_slm execCommand $event $list_args]

	    return $return_exec_command
		
	} else {

	    return $return_apply_event

	}
    }

     method execCommandExp { slm_id cmd list_args transaction_id } {

	puts "\nMethod execCommandExp of instrument server for $slm_id $cmd $list_args\n"

	set obj_slm [[self] getObjModule $slm_id]

	return [$obj_slm execCommandExp $cmd $list_args $transaction_id ]
	
    }
}




###############################################################
#Instrument module
###############################################################

oo::class create instrumentModule {

    variable inst_module_attributes
    
    constructor { name } {

	dict set inst_module_attributes name $name
    }

    method getID {} {

	#puts "Method getID of [self]\n"
	
	return [dict get $inst_module_attributes name]
    }

    

    method setState { state } {

	#puts "Method setState of [self] for state -> $state\n"
	
	dict set inst_module_attributes state $state

    }

    #Rather than recovering the name from the class name,
    #a command could also be implemented in each state's entryAction method
    #to reset the state name
    
    method getState {} {

	#puts "Method getState of [self]\n"

	set obj_state [dict get $inst_module_attributes state]

	return [ string range [info object class $obj_state] 7 end ]

    }

    method setDataFlowController { dfc } {

	puts "\nMethod setDataFlowController of [self], from [info object class [self]] class\n"
	
	dict set inst_module_attributes dfc $dfc

    }

    
    method loadSLM {} {

	#puts "Method loadSLM [[self] getID] of [self]\n"

	#Load the driver to the instance of equipment

	#https://stackoverflow.com/questions/33415534/find-all-procs-of-a-namespace-in-tcl
	#puts " **The procedures BEFORE loading: [info procs ::[[self] getID]::* ] **\n"
	package require [[self] getID]
	
	#puts " **The procedures AFTER loading: [info procs ::[[self] getID]::* ] **\n"

	set list_procedures [info procs ::[[self] getID]::* ]

	set list_commands {}
	
	foreach p $list_procedures {
	    
	    #puts "Procedure [namespace tail $p]\n"
	    lappend list_commands [namespace tail $p]

	}

	return $list_commands
    }

    method getArgsCommandSLM { command_name } {

	#Returns the names of the arguments for a procedure
	return [info args ::[[self] getID]::$command_name]
    }

    method execCommand { cmd list_args } {
	
	set timeout 200
	set vwait_var 1
	
	set transaction_id [ clock milliseconds ]
	
	puts "Instrument module [self], at state [[self] getState], executes the commmand $cmd with the args $list_args\n"
	
	set serial_controller [dict get $inst_module_attributes dfc]
	
	if { $list_args != "NO_ARG" } {
	    
	    return [::[[self] getID]::$cmd $serial_controller $transaction_id {*}$list_args]
	    
	} else {
	    
	    for { set i 0 } { $i < 10 } { incr i } {
		
		::[[self] getID]::$cmd $serial_controller $transaction_id
		
		after $timeout [list set vwait_var [expr $vwait_var * -1] ]
		
		vwait vwait_var
		
		puts "Method execCommand i: $i"
		
		set command_return [$serial_controller getData $transaction_id]
		
		puts "\nMethod execCommand command_return: $command_return\n"
		
		if { $command_return == "ACK" } {
		    
		    return "ACK"
		    
		} 
	    }

	    return "NACK"
	    
	}
	
    }
    
    method applyEvent { event } {

	set obj_state [dict get $inst_module_attributes state]

	set return_apply_event [$obj_state applyStateEvent $event]

	return $return_apply_event
    }

    method execCommandExp { cmd list_args transaction_id } {

	puts "Method execCommandExp - instrument module [self], at state [[self] getState], executes the commmand $cmd with the args $list_args\n"
	 
	set serial_controller [dict get $inst_module_attributes dfc]

	::[[self] getID]::setInternalArgs $serial_controller $transaction_id
	
	set return_change_state [[self] applyEvent run_op]

	if { $return_change_state == "ACK" } {
	    
	    if { $list_args != "NO_ARG" } {
		
		if { [catch { ::[[self] getID]::$cmd {*}$list_args } result] != 0 } {

		    [self] applyEvent op_denied
		    
		    return "NACK;$result"
		    
		} else {
		    
		    [self] applyEvent op_started
		    [self] applyEvent op_completed
		    
		    return "ACK;$result"
		    
		}
		
	    } else {
		
		
		if { [catch { [::[[self] getID]::$cmd ] } result] != 0 } {

		    [self] applyEvent op_denied
		    
		    return "NACK;$result"
		    
		} else {
		    
		    [self] applyEvent op_started
		    [self] applyEvent op_completed

		    return "ACK;$result"
		}
		
	    }

	} else {

	    return "NACK;$return_change_state"

	}
    }
}

###############################################################
#Equipment server
###############################################################

oo::class create equipmentServer {

    variable equip_server_attributes
    
    constructor { } {

    }

    method setAvailableEquip { available_resources } {

	#puts "Method setAvailableEquip of [self]"

	foreach slm [dict keys $available_resources] {

	    set t [dict get $available_resources $slm type]

	    if { $t == "equipment" } {

		set n [dict get $available_resources $slm name]
		dict set equip_server_attributes equip $slm name $n

	    }

	}

    }

    method getAvailableEquip {} {

	#puts "Method getAvailableEquip returns [dict get $equip_server_attributes equip]"

	return [dict get $equip_server_attributes equip]

    }

    method setDataFlowController { dfc } {

	dict set equip_server_attributes dfc $dfc
	
    }

    method getDataFlowController {} {


	return [dict get $equip_server_attributes dfc]

    }
    
    method createEquipModule { slm_id } {

	#puts "Method createEquipModule of [self] for slm:$slm_id"

	set obj_active_slm [equipmentModule new $slm_id]

	set state_obj_active_slm [[self] createEquipState $obj_active_slm]

	$obj_active_slm setState $state_obj_active_slm

	#Agregate de instance of dataFlowController to the equipmentModule
	set data_flow_controller [[self] getDataFlowController]

	$obj_active_slm setDataFlowController $data_flow_controller
	
	dict lappend equip_server_attributes obj_active_slm $obj_active_slm
	
	return $obj_active_slm
	       
    }

    #This method receive the name of the object of module (::oo::Objnn)
    #and returns the corresponding id (slm_nn)
    method getIDModule { slm_obj } {

	return [$slm_obj getID]

    }

    #This method receive the id of slm (slm_nn) and returns
    #the corresponding name of the object of module (::oo::Objnn)
    method getObjModule { slm_id } {
	
	#puts "[self] - method getObjModule for $slm_id\n"
	
	foreach obj_slm [[self] getSelectedObjSLM] {
	    
	    if { [[self] getIDModule $obj_slm] == $slm_id } {
		
		return $obj_slm
		
	    } 
	    
	}
	
	return 0
	
    }
	
    method createEquipState { id_obj } {

	#puts "Method createEquipState of [self] for id_obj -> $id_obj\n"
	
	set obj_state [stateUNCONNECTED new $id_obj]

	return $obj_state
	
    }

    method getSelectedObjSLM {} {

	#puts "Method getSelectedObjSLM of [self]"

	if { [ dict exists $equip_server_attributes obj_active_slm ] } {
	    return [ dict get $equip_server_attributes obj_active_slm ]
	} else {
	    return 0
	}

    }

    method setAvailableCommands { available_resources } {

	#puts "\nMethod setAvailableCommands of equipmentServer ([self])\n"

	foreach active_obj_slm [[self] getSelectedObjSLM] {

	    if { $active_obj_slm != 0 } {
		
		#puts "SLM of $active_obj_slm -> [$active_obj_slm getID]"
		
		set slm_active [$active_obj_slm getID]
		
		foreach slm [dict keys $available_resources] {
		    
		    #puts "Loop foreach with slm -> $slm and slm_active -> $slm_active\n"
		    
		    if { $slm == $slm_active } {
			
			set c [dict get $available_resources $slm command]
			
			#puts "Commands available of equipment $slm [dict keys $c]\n"
			
			dict set equip_server_attributes $slm command $c
			
		    }
		    
		}
		
	    }
	}

    }
    
    method getAvailableCommands {} {
	
	#puts "Method getAvailableCommands of equipmentServer [self]\n"

	foreach active_obj_slm [[self] getSelectedObjSLM] {

	    if { $active_obj_slm != 0 } {
		
		set slm [$active_obj_slm getID]

		#puts "$slm [dict get $equip_server_attributes $slm command]"

		dict set commands_equip $slm [dict get $equip_server_attributes $slm command]
	    }
	}

	if { [info exists commands_equip ] } {
	    return $commands_equip
	} else {
	    return ""
	}
    }

    method startConnection { slm_id } {
	
	puts "\n# Method startConnection of [self] ([info object class [self]]) for slm_id - $slm_id\n"

		set data_flow_controller [[self] getDataFlowController]

	set id_serial_port [$data_flow_controller getSerialPortID]

	puts "\nThe dataFlowController $data_flow_controller and de id_serial_port $id_serial_port\n"

	if { ($id_serial_port == "0") } {

	    $data_flow_controller openPort

	    set id_serial_port [$data_flow_controller getSerialPortID]
	    
	    puts "\nAfter openPort id_serial_port $id_serial_port\n"
	}
	
	set obj_active_slm [[self] createEquipModule $slm_id]

	set return_start_connection [$obj_active_slm applyEvent eventSTART_CONNECTION]

	#puts "\nreturn_start_connection - $return_start_connection\n"

	if { $return_start_connection == "ACK" } {
	    
	    [self] loadCommands $obj_active_slm

	    set return_check_connection [[self] execCommand $obj_active_slm checkConnection NO_ARG]
	    #puts "return_check_connection - $return_check_connection"

	    if { $return_check_connection == "ACK" } {

		$obj_active_slm applyEvent eventCONNECTION_ACCEPTED

	    }
	    
	    return $return_check_connection 
	    
	} else {


	    return $return_start_connection

	}
	
    }
    
    method setLoadedCommands { id_obj_slm list_commands } {

	#puts " ****Equipment Server - Method setLoadedCommands ob_slm: $id_obj_slm commands: $list_commands **** \n"

	dict set equip_server_attributes loaded_commands $id_obj_slm $list_commands

    }

    method getLoadedCommands { id_obj_slm } {

	return [dict get $equip_server_attributes loaded_commands $id_obj_slm]

    }

    method loadCommands { obj_slm } {

	set list_commands [$obj_slm loadSLM]

	[self] setLoadedCommands $obj_slm $list_commands

	#puts " ***Equipment Server - Method loadCommands of slm $obj_slm load the commands $list_commands ***\n"
	
    }

    method getArgsCommand { obj_slm command_name } {

	puts "Method getArgsCommand with obj_slm:$obj_slm and command_name:$command_name"
	
	set list_args_command [$obj_slm getArgsCommandSLM $command_name]

	return $list_args_command
    }

    method execCommand { obj_slm cmd list_args} {

	#puts "Method execCommand of instrumentServer [self] module:$obj_slm cmd:$cmd args:$list_args"

	return [$obj_slm execCommand $cmd $list_args]
	
    }

    method applyEventSLM { slm_id event list_args} {

	puts "Method applyEventSLM of [info object class [self]] with $slm_id $event and $list_args\n"

	set obj_slm [[self] getObjModule $slm_id]
	
	set return_apply_event [ $obj_slm applyEvent $event]

	puts "Return of apply $event in module [$obj_slm getID] is $return_apply_event\n"

	if { $return_apply_event == "ACK" } {

	    set return_exec_command [$obj_slm execCommand $event $list_args]

	    return $return_exec_command

	} else {

	    return $return_apply_event
	}
    }

    method execCommandExp { slm_id cmd list_args transaction_id } {

	puts "\nMethod execCommandExp of equipment server for $slm_id $cmd $list_args $transaction_id\n"

	set obj_slm [[self] getObjModule $slm_id]

	return [$obj_slm execCommandExp $cmd $list_args $transaction_id]
	
    }
}


###############################################################
#Equipment module
###############################################################

oo::class create equipmentModule {

    variable equip_module_attributes
    
    constructor { name } {

	dict set equip_module_attributes name $name
    }

    method getID {} {

	#puts "Method getID of [self]\n"

	return [dict get $equip_module_attributes name]

    }

    method setState { state } {

	#puts "Method setState of [self] for state -> $state\n"

	dict set equip_module_attributes state $state

    }

    method getState {} {

	#puts "Method getState of [self]\n"

	set obj_state [dict get $equip_module_attributes state]

	return [ string range [info object class $obj_state] 7 end ]

    }

    method setDataFlowController { dfc } {

	puts "\nMethod setDataFlowController of [self], from [info object class [self]] class\n"
	
	dict set equip_module_attributes dfc $dfc

    }

    method loadSLM {} {

	#puts "Method loadSLM [[self] getID] of [self]\n"

	#Load the driver to the instance of equipment

	#https://stackoverflow.com/questions/33415534/find-all-procs-of-a-namespace-in-tcl
	#puts " **The procedures BEFORE loading: [info procs ::[[self] getID]::* ] **\n"
	package require [[self] getID]

	#puts " **The procedures AFTER loading: [info procs ::[[self] getID]::* ] **\n"
	
	set list_procedures [info procs ::[[self] getID]::* ]

	set list_commands {}
	
	foreach p $list_procedures {
	    
	    #puts "Procedure [namespace tail $p]\n"
	    
	    lappend list_commands [namespace tail $p]

	}

	return $list_commands
    }

    method getArgsCommandSLM { command_name } {

	puts "Method gerArgsCommandSLM: [info args ::[[self] getID]::$command_name]"
	
	return [info args ::[[self] getID]::$command_name]
    }

    #Method to be used during connection
    method execCommand { cmd list_args } {

	set timeout 200
	set vwait_var 1
	
	set transaction_id [ clock milliseconds ]

	puts "Equipment module [self], at state [[self] getState], executes the commmand $cmd with the args $list_args"

	set serial_controller [dict get $equip_module_attributes dfc]
	
	if { $list_args != "NO_ARG" } {
	    
	    return [::[[self] getID]::$cmd $serial_controller $transaction_id {*}$list_args]

	} else {

	    for { set i 0 } { $i < 10 } { incr i } {
		
		::[[self] getID]::$cmd $serial_controller $transaction_id
		
		after $timeout [list set vwait_var [expr $vwait_var * -1] ]
		
		vwait vwait_var
		
		puts "Method execCommand i: $i"
		
		set command_return [$serial_controller getData $transaction_id]
		
		puts "\nMethod execCommand command_return: $command_return\n"
		
		if { $command_return == "ACK" } {
		    
		    return "ACK"
		    
		} 
	    }

	    return "NACK"

	}
    }

    method applyEvent { event } {

	set obj_state [dict get $equip_module_attributes state]

	set return_apply_event [$obj_state applyStateEvent $event]

	return $return_apply_event
    }

    #Method to exec commands during an experiment
    method execCommandExp { cmd list_args transaction_id } {

	puts "Method execCommandExp - equipment module [self], at state [[self] getState], executes the commmand $cmd with the args $list_args and transaction_id: $transaction_id\n"

	set serial_controller [dict get $equip_module_attributes dfc]

	#Call the procedure setInternalArgs in the namespace of current slm module (::[[self] getID]::)
	#and pass the id of serial_controller and transaction_id
	
	::[[self] getID]::setInternalArgs $serial_controller $transaction_id
	
	set return_change_state [[self] applyEvent run_op]

	if { $return_change_state == "ACK" } {
	    
	    if { $list_args != "NO_ARG" } {
		
		if { [catch { ::[[self] getID]::$cmd {*}$list_args } result] != 0 } {
		    
		    [self] applyEvent op_denied

		    return "NACK;$result"
		    
		} else {
		    
		    [self] applyEvent op_started
		    [self] applyEvent op_completed
		    
		    return "ACK;$result"
		    
		}
		
	    } else {
		
		
		if { [catch { [::[[self] getID]::$cmd] } result] != 0 } {
		    
		    [self] applyEvent op_denied

		    return "NACK;$result"
		    
		} else {
		    
		    [self] applyEvent op_started
		    [self] applyEvent op_completed

		    return "ACK;$result"
		    
		}
		
	    }

	} else {

	    return "NACK;$return_change_state"

	}
    }
    

}


###############################################################
#Info about Machine State Object Oriented
###############################################################

#http://www.conradbock.org/statemachine.html
#Object-oriented States
#This technique starts by treating states as classes that are instantiated
#and deleted at runtime as the state machine executes,

#https://www.tclcommunityassociation.org/wub/proceedings/Proceedings-2012/SeanWoods/Lifecycle-of-Objects.pdf
#Lifecycle Object  Generators (LOG)

###############################################################
#State UNCONNECTED
###############################################################

oo::class create stateUNCONNECTED {

    variable slm_state_attributes
    
    constructor { name } {

	dict set slm_state_attributes name_object $name
    }

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method eventSTART_CONNECTION {} {
	
	#Method exitAction is defined in the current state
	[self] exitAction
	oo::objdefine [self] class stateCONNECTING
	#Method entryAction is defined in the next state
	#and the return send info to the caller
	[self] entryAction
	
    }

    method entryAction {} {

	puts "Object [self] entering state current [info object class [self object]]"

    }

    method exitAction {} {

	puts "Object [self] leaving the current state [info object class [self]]"

    }

    method getName {} {

	return [dict get $slm_state_attributes name_object]
    }

}

###############################################################
#State CONNECTING
###############################################################

oo::class create stateCONNECTING {

    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method eventCONNECTION_ACCEPTED {} {

	#Method exitAction is defined in the current state
	[self] exitAction

	oo::objdefine [self] class stateCONNECTED

	#Method exitAction is defined in the current state
	[self] entryAction

	[self] exitAction
	
	oo::objdefine [self] class stateOPERATING

	[self] entryAction

	[self] exitAction
	
	oo::objdefine [self] class stateCONTROL_FLOW

	[self] entryAction

	[self] exitAction

	oo::objdefine [self] class statePOWERED_UP

	[self] entryAction
	
    }

    method eventERROR_CONNECTION {} {
	
	[self] exitAction
	
	oo::objdefine [self] class stateUNCONNECTED
	
	[self] entryAction
    }    

    method entryAction {} {

	puts "Object [self] entering the current state [info object class [self object]]"

    }

    method exitAction {} {

	puts "Object [self] leaving state current [info object class [self]]"

    }
    
    method getName {} {

	return [dict get $slm_state_attributes name_object]
    }

}

###############################################################
#State CONNECTED
###############################################################

oo::class create stateCONNECTED {

    variable slm_state_attributes
    
    constructor { name } {

    }

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method eventSTART_DISCONNECTION {} {

	[self] exitAction
	
	oo::objdefine [self] class stateDISCONNECTING

	[self] entryAction
	
    }

    method getName {} {

	return [dict get $slm_state_attributes name_object]
    }

    method entryAction {} {

	puts "Object [self] entering state current [info object class [self]]"

    }

    method exitAction {} {

	puts "Leaving state current [info object class [self]]"

    }

}

###############################################################
#State DISCONNECTING
###############################################################

oo::class create stateDISCONNECTING {

    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method eventEND_DISCONNECTION {} {

	[self] exitAction
	
	oo::objdefine [self] class stateUNCONNECTED

	[self] entryAction
    }

    method getName {} {

	return [dict get $slm_state_attributes name_object]
    }

     method entryAction {} {

	#puts "[self] entering state current [info object class [self]]" 

     }

    method exitAction {} {
	     
	    #puts "Leaving state current [info object class [self]]"
	     
    }
}

###############################################################
#State OPERATING
###############################################################

oo::class create stateOPERATING {

    superclass stateCONNECTED
    
    variable slm_state_attributes

    constructor { name } {

    }

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method entryAction {} {

	puts "Object [self] entering state current [info object class [self]]"

		
    }
    
    method exitAction {} {

	puts "Object [self] leaving state current [info object class [self]]"

    }

}

###############################################################
#State CONTROL_FLOW
###############################################################


oo::class create stateCONTROL_FLOW {

    superclass stateOPERATING
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method entryAction {} {

	puts "Object [self] entering state current [info object class [self]]"

		
    }
    
    method exitAction {} {

	puts "Object [self] leaving state current [info object class [self]]"

    }
}

    
###############################################################
#State POWERED_UP
###############################################################

oo::class create statePOWERED_UP {

    superclass stateCONTROL_FLOW
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method entryAction {} {

	puts "Object [self] entering state current [info object class [self]]"

		
    }
    
    method exitAction {} {

	puts "Object [self] leaving state current [info object class [self]]"

    }

    method init {} {

	puts "Method eventINIT of [self]"

	[self] exitAction

	oo::objdefine [self] class stateINITING

	[self] entryAction
	
    }
}

###############################################################
#State INITING
###############################################################

oo::class create stateINITING {

    superclass stateCONTROL_FLOW
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method entryAction {} {
	
	puts "Object [self] entering state current [info object class [self]]"
	
	
    }
    
    method exitAction {} {
	
	puts "Object [self] leaving state current [info object class [self]]"
	
    }

    method end_init {} {

	[self] exitAction

	oo::objdefine [self] class stateIDLE

	[self] entryAction
    }

    method error_init {} {

	[self] exitAction

	oo::objdefine [self] class statePOWERED_UP

	[self] entryAction
    }
}


###############################################################
#State IDLE
###############################################################

oo::class create stateIDLE {
    
    superclass stateCONTROL_FLOW
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {
	
	puts "Object [self] entering state current [info object class [self]]"
	
	
    }
    
    method exitAction {} {
	
	puts "Object [self] leaving state current [info object class [self]]"
	
    }
    
    method setup {} {
	
	puts "Method eventSetup of [self] at current state [info object class [self]]"
	
	[self] exitAction
	
	oo::objdefine [self] class stateCONFIGURING
	
	[self] entryAction
	
    }

}

###############################################################
#State CONFIGURING
###############################################################

oo::class create stateCONFIGURING {
    
    superclass stateCONTROL_FLOW
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {
	
	puts "Object [self] entering state current [info object class [self]]"
	
	
    }
    
    method exitAction {} {
	
	puts "Object [self] leaving state current [info object class [self]]"
	
    }

    method end_setup {} {

	[self] exitAction

	oo::objdefine [self] class stateNORMAL_OPERATION

	[self] entryAction
	
    }

    method error_setup {} {

	[self] exitAction

	oo::objdefine [self] class stateIDLE

	[self] entryAction
	
    }

    

}

###############################################################
#State NORMAL_OPERATION
###############################################################

oo::class create stateNORMAL_OPERATION {
    
    superclass stateCONTROL_FLOW
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {
	
	puts "Object [self] entering state current [info object class [self]]"
	
	
    }
    
    method exitAction {} {
	
	puts "Object [self] leaving state current [info object class [self]]"
	
    }

    method run_op { } {

	puts "Method eventRunOP of [self]\n"

	[self] exitAction
	
	oo::objdefine [self] class statePROCESSING_REQUESTED
	
	[self] entryAction

    }

    method eventCLEAR { } {

	puts "Method eventClear of [self]\n"

	[self] exitAction
	
	oo::objdefine [self] class stateCLEARING
	
	[self] entryAction

	[self] exitAction

	oo::objdefine [self] class stateIDLE

	[self] entryAction

	
    }

}

###############################################################
#State PROCESSING_REQUESTED
###############################################################

oo::class create statePROCESSING_REQUESTED {
    
    superclass stateCONTROL_FLOW
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {
	
	puts "Object [self] entering state current [info object class [self]]"
	
	
    }
    
    method exitAction {} {
	
	puts "Object [self] leaving state current [info object class [self]]"
	
    }

    method op_started { } {

	puts "Method eventOP_STARTED of [self]\n"

	[self] exitAction
	
	oo::objdefine [self] class statePROCESSING
	
	[self] entryAction
	
    }

    method op_denied { } {

	puts "Method eventOP_DENIED of [self]\n"

	[self] exitAction
	
	oo::objdefine [self] class stateTERMINATED
	
	[self] entryAction

	[self] exitAction
	
	oo::objdefine [self] class stateNORMAL_OPERATION
	
	[self] entryAction
	
    }

}

###############################################################
#State PROCESSING
###############################################################

oo::class create statePROCESSING {
    
    superclass stateCONTROL_FLOW
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {
	
	puts "Object [self] entering state current [info object class [self]]"
	
	
    }
    
    method exitAction {} {
	
	puts "Object [self] leaving state current [info object class [self]]"
	
    }

    method eventOP_RESULT { result } {

	puts "Method eventOP_RESULT of [self] with result:$result\n"
	
	
    }

    method op_completed { } {

	puts "Method eventOP_COMPLETED of [self]\n"

	[self] exitAction
	
	oo::objdefine [self] class stateTERMINATED
	
	[self] entryAction

	[self] exitAction
	
	oo::objdefine [self] class stateNORMAL_OPERATION
	
	[self] entryAction
	
    }

    method eventALARM_ON { } {

	puts "Method eventALARM_ON of [self]\n"

	[self] exitAction
	
	oo::objdefine [self] class stateALARM
	
	[self] entryAction
	
    }

}

###############################################################
#State TERMINATED
###############################################################

oo::class create stateTERMINATED {
    
    superclass stateCONTROL_FLOW
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {
	
	puts "Object [self] entering state current [info object class [self]]"
	
	
    }
    
    method exitAction {} {
	
	puts "Object [self] leaving state current [info object class [self]]"
	
    }

}

###############################################################
#State ALARM
###############################################################

oo::class create stateALARM {
    
    superclass stateCONTROL_FLOW
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {
	
	puts "Object [self] entering state current [info object class [self]]"
	
	
    }
    
    method exitAction {} {
	
	puts "Object [self] leaving state current [info object class [self]]"
	
    }

    method eventALARM_OFF { } {

	puts "Method eventALARM_OFF of [self]\n"

	[self] exitAction
	
	oo::objdefine [self] class statePROCESSING
	
	[self] entryAction
	
    }

}

###############################################################
#State CLEARING
###############################################################

oo::class create stateCLEARING {
    
    superclass stateCONTROL_FLOW
    
    variable slm_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {
	
	puts "Object [self] entering state current [info object class [self]]"
	
	
    }
    
    method exitAction {} {
	
	puts "Object [self] leaving state current [info object class [self]]"
	
    }

}



###############################################################
#Class dataFlowController
###############################################################

oo::class create dataFlowController {
    
    variable data_flow_controller_attributes

    constructor {} {

	puts "Constructor of dataFlowController"

    }

    method requestSendCommand { cmd } {

    }

    method checkCommandList {} {

    }
    
    method sendCommand { cmd } {

	set id_serial_port [dict get $data_flow_controller_attributes id_serial_port]

	puts "\nMethod sendCommand of [self] ([info object class [self]]) sending cmd: $cmd to id_serial_port $id_serial_port\n"

	if { [chan eof $id_serial_port] } {

	    puts "Method sendCommand: SERIAL_PORT_CLOSED"

	   return "SERIAL_PORT_CLOSED"

	} else {

	    puts "Method sendCommand: Serial Port Open"
	    
	    chan puts $id_serial_port $cmd

	    return "COMMAND_SENT"
	    
	}
	
    }

    method openPort {} {

	
	puts "Method openPort of [self]"
	
	#https://www.tcl.tk/man/tcl8.5/tutorial/Tcl26.html
	if { [catch {set serial_device [exec ls {*}[glob /dev/ttyACM*]]} ] } {

	    dict set  data_flow_controller_attributes serial_device $serial_device
	    
	    return "NO_SERIAL_DEVICE"
	    
	} else {
	    
	    dict set  data_flow_controller_attributes serial_device $serial_device
	    
	    set baud 9600
	    set parity n
	    set data_bit 8
	    set stop_bit 1
	    
	    set result_1 [ catch { set id_serial_port [open $serial_device r+] } ]
	    
	    set result_2 [catch {fconfigure $id_serial_port -mode "$baud,$parity,$data_bit,$stop_bit"}]
	    
	    if { ($result_1 != 0) || ($result_2 != 0) } {
		
		set id_serial_port "SERIAL_ERROR_OPEN"
		
		dict set data_flow_controller_attributes id_serial_port $id_serial_port
		
		return $id_serial_port
		
	    } else {
		
		dict set data_flow_controller_attributes id_serial_port $id_serial_port
		
		chan configure $id_serial_port -buffering line

		#Worked with crlf crlf but worked also without set translation
		#so wi commented
		#chan configure $id_serial_port -translation { crlf crlf }

		chan configure $id_serial_port -blocking 0

		chan configure $id_serial_port -encoding ascii

		#chan configure $id_serial_port -encoding utf-8

		#file event handler to be called whenever
		#id_serial_port enters the state described by event
		#(which must be either readable or writable ).

		#chan event $id_serial_port readable [list [self] receiveReadout]
		#fileevent $id_serial_port readable [list [self] receiveReadout $id_serial_port]

		chan event $id_serial_port readable [list [self] receiveReadout $id_serial_port]
		
		#chan event $id_serial_port readable [list read_port $id_serial_port]
		
		#To show the configuration parameters of id_serial_port
		#puts "\n[chan configure $id_serial_port]\n"

		#To show the command to be called when id_serial_por is readable
		#puts "\nFileevent inside openPort: [chan event $id_serial_port readable]\n"
		
		return $id_serial_port
		
	    }
	    
	}

    }
	
	
    method getSerialPortID {} {

	puts "\nMethod getSerialPortID\n"

	if { [ info exists data_flow_controller_attributes] && [dict exists  $data_flow_controller_attributes id_serial_port ] } {
	    return [dict get $data_flow_controller_attributes id_serial_port ]
	} else {
	    return 0
	}
	
    }
    
    method closePort { } {


    }

    #Receives the data sent by serial port and stores in
    #data_flow_controller_attributes
    
    method receiveReadout { id_serial_port } {

	puts "\n ##### Method receiveReadout from $id_serial_port  ##### \n"
	
	if { [eof $id_serial_port ] } {
	    puts "Closing $id_serial_port"
	    catch { close $id_serial_port }
	    return
	}

	
	set serial_return [chan gets $id_serial_port]

	if {$serial_return == ""} {
	    
	    return

	} else {

	    [self] saveReadout $serial_return
	}

	puts "receiveReadout - serial return: $serial_return\n"

    }

    method saveReadout { readout } {

	puts "\nMethod saveReadout readout $readout\n
"
	set readout_list [split $readout ";"]

	set transaction_id [lindex $readout_list end]

	set readout [lrange $readout_list 0 end-1]

	if { [dict exists $data_flow_controller_attributes received_data $transaction_id ] } {

	    dict with data_flow_controller_attributes received_data {

		lappend $transaction_id $readout
		
	    }
	    
	} else {

	    dict set data_flow_controller_attributes received_data $transaction_id $readout

	}

	puts "\nsaveReadout readout: $readout transaction_id: $transaction_id and $data_flow_controller_attributes\n"
	
    }

    method getData { transaction_id } {

	puts "\nBEGIN getData: transaction_id: $transaction_id and $data_flow_controller_attributes\n"
	
	if { [dict exists $data_flow_controller_attributes received_data $transaction_id ] } {

	    #Get the data with key = transaction_id
	    set data_received [dict get $data_flow_controller_attributes received_data $transaction_id ]

	    #Unset the key transaction_id
	    dict unset  data_flow_controller_attributes received_data $transaction_id

	    puts "\nEND getData: data_received $data_received  $data_flow_controller_attributes\n"

	    return $data_received

	} else {

	    puts "\nEND getData: NO_DATA  $data_flow_controller_attributes\n"
	    
	    return "NO_DATA"

	}

    }
    

}

