#!/bin/sh
#A próxima linha reinicia usando o wish \
	exec wish "$0" "$@"



#    "Copyright (C) 2021  Info Libera (liberinforma[at]gmail.com)"

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#30/06/2021
#Version 00
#User selection of SLMs to be used
#Loading of SLMs
#User programming of experiment
#Validation of program

#30/07/2021
#Version 01
#Implement State Machine of SLMs
#Implement State Machine of TSC

#17/08/2021
#Version 03
#Implement execCommandExp to execute commands during an experiment
#and the events from NORMAL_OPERATION state
#Implement the class dataFlowController

#25/08/2021
#Version 04
#Implement code to run an experiment
#Method waitTime for TSC

#14/12/2021
#SisProLab - Version 00

#21/12/2021
#SisProLab - Version 01
#Include slm_04 (syringe pump)

###############################################################
#Load the file with the definition of classes
###############################################################

set classes_file sisprolab_class_01.tcl

source $classes_file

puts "Loaded $classes_file"


###############################################################
#Controller of the experiment
###############################################################

oo::class create experimentController {

    variable exp_controller_attributes
    
    constructor { } {

    }

    method applyEventTSC { event } {

	set obj_state [dict get $exp_controller_attributes obj_state]

	set return_apply_event [$obj_state applyStateEvent $event]

	[self] updateStatusExperiment
	
	#puts "Return of event $event $return_apply_event"

	return $return_apply_event

    }
    
    #Command to be only just in UNCONFIGURED state
    method connectInstServer { obj_id } {

	set obj_state [dict get $exp_controller_attributes obj_state]

	set return_check_command [$obj_state checkCommandTSC [self method]]

	if {  $return_check_command == "ACK" } {
	    
	   #puts "Method [self method]: command executed: $return_check_command"
	    
	    #dict lappend exp_controller_attributes server $obj_id $type
	    
	    dict set exp_controller_attributes inst obj_id $obj_id
	    
	} else {

	   #puts "Method [self method]: command NOT executed: $return_check_command"

	}
    }

    #Command to be only just in UNCONFIGURED state
    method connectEquipServer { obj_id } {

	set obj_state [dict get $exp_controller_attributes obj_state]

	set return_check_command [$obj_state checkCommandTSC [self method] ]

	if {  $return_check_command == "ACK" } {
	    
	   #puts "Method [self method]: command executed: $return_check_command"
	    
	    dict set exp_controller_attributes equip obj_id $obj_id
	    
	} else {
	    
	   #puts "Method [self method]: command NOT executed: $return_check_command"
	    
	}
	    

    }

    #Command to be only just in UNCONFIGURED state
    method setObjState { obj_state } {


	if { [ info exists exp_controller_attributes] && [ dict exists $exp_controller_attributes obj_state] } {
	    
	    set obj_state [dict get $exp_controller_attributes obj_state]

	} else {

	    set obj_state $obj_state

	}

	set return_check_command [$obj_state checkCommandTSC [self method] ]

	if {  $return_check_command == "ACK" } {
	    
	   #puts "Method [self method]: command executed: $return_check_command"
	    
	   #puts "Method [self method] of [self] aggregates $obj_state (state: [info object class $obj_state]) to [self]"
	    
	    dict set exp_controller_attributes obj_state $obj_state
	    
	} else {
	    
	   #puts "Method [self method]: command NOT executed: $return_check_command"
	    
	}

	    
    }
    
    method getServer { type } {

	#puts "Method getServer of [self]  for type:$type\n"

	#Type may be inst or equip and returns the id of the obj

	if { [ dict exists $exp_controller_attributes $type obj_id ] } {
	    
	    return [ dict get $exp_controller_attributes $type obj_id]

	} else {

	    return 0

	}

    }


    #Transfer the info about the available SLMs to the experiment controller
    #$experiment_controller setAvailableSLM [$equipment_server getAvailableEquip]

    method setAvailableSLM { slm } {

	#puts " *Method setAvailableSLM of [self] $slm\n"

	foreach key_slm [dict keys $slm] {

	    dict lappend exp_controller_attributes available_resources $key_slm

	}
	
    }

    method getAvailableSLM {} {

	if { [dict exists $exp_controller_attributes available_resources] } {
	    return [dict get $exp_controller_attributes available_resources]
	} else {
	    return 0
	}

    }

    method getAvailableInstSLM {} {

	set inst_server [[self] getServer inst]
	set inst_available_resources [$inst_server getAvailableInst]
	return $inst_available_resources
	
    }

    method getAvailableEquipSLM {} {

	set equip_server [[self] getServer equip]
	set equip_available_resources [$equip_server getAvailableEquip]
	return $equip_available_resources
	
    }

    method setSelectedInstSLM { slm_id } {

	#puts "Method setSelectedInstSLM of [self] for SLM $slm_id\n"
	
	dict lappend exp_controller_attributes selected_inst_slm $slm_id

    }

    method setSelectedEquipSLM { slm_id } {

	#puts "Method setSelectedEquipSLM of [self] for SLM $slm_id\n"

	dict lappend exp_controller_attributes selected_equip_slm $slm_id

    }

    #Returns the list of instruments selected by the user
    #But not necessarily active
    #Will be active if there is no error in the connection step
    method getSelectedInstSLM {} {

	if { [dict exists $exp_controller_attributes selected_inst_slm] } {
	    return [dict get $exp_controller_attributes selected_inst_slm]
	} else {
	    return 0
	}

    }

    #Returns the list of equipments selected by the user
    #But not necessarily active
    #Will be active if there is no error in the connection step
    method getSelectedEquipSLM {} {

	if { [dict exists $exp_controller_attributes selected_equip_slm] } {
	    return [dict get $exp_controller_attributes selected_equip_slm]
	} else {
	    return 0
	}
    }

    method setActiveInstSLM { slm_id } {

	#puts "Method setActiveInstSLM of [self] for SLM $slm_id\n"
	
	dict lappend exp_controller_attributes active_inst_slm $slm_id

    }

    method setActiveEquipSLM { slm_id } {

	#puts "Method setActiveEquipSLM of [self] for SLM $slm_id\n"
	
	dict lappend exp_controller_attributes active_equip_slm $slm_id

    }

    method getActiveInstSLM {} {

	if { [dict exists $exp_controller_attributes active_inst_slm] } {
	    return [dict get $exp_controller_attributes active_inst_slm]
	} else {
	    return 0
	}

    }

    method getActiveEquipSLM {} {

	if { [dict exists $exp_controller_attributes active_equip_slm] } {
	    return [dict get $exp_controller_attributes active_equip_slm]
	} else {
	    return 0
	}
    }
        
    method setAllAvailableCommands { available_resources } {

	#puts " ****Method setAllAvailableCommands available_resources: $available_resources\n"
	
	set inst_server [[self] getServer inst]

	$inst_server setAvailableCommands $available_resources

	dict set exp_controller_attributes inst_commands \
	    [$inst_server getAvailableCommands]
	
	set equip_server [[self] getServer equip]

	$equip_server setAvailableCommands $available_resources

	dict set exp_controller_attributes equip_commands \
	    [$equip_server getAvailableCommands]
	    
    }

    method getAllAvailableCommands { typeSLM_commands } {

	return [dict get $exp_controller_attributes $typeSLM_commands]
	
    }

    method validateProgram { program_window log_window } {
	
	set return_event [[self] applyEventTSC eventVALIDATE_PROGRAM]

	#Tip - http://pages.cpsc.ucalgary.ca/~saul/personal/archives/Tcl-Tk_stuff/tcl_examples/
	#Create a tag to color alert messages
	$log_window tag configure alert_message -foreground red
	
	if { $return_event  == "ACK" } {
	    
	    set status_program ACK
	    
	    #puts "Method validateProgram of [self] using the program window $program_window"
	    
	    #Clean the window of log
	    $log_window delete 1.0 end
	    
	    set program_to_run [$program_window get 1.0 end]
	    
	    #puts "Program to run -> $program_to_run"
	    
	    set program_list [split $program_to_run "\n"]
	    
	    set i 0
	    set program_to_run_list {}
	    
	    foreach line $program_list {
		
		#Test if:
		# not an empty element ($line != "")
		#(because the last element of list is empty
		# not a comment ![regexp {^#.*} $line]
		# not new line ![regexp {^\n} $line]
		# non-whitespace [regexp {\S+} $line]
		
		if {($line != "") && \
			![regexp {^#.*} $line] && \
			![regexp {^\n} $line] && \
			[regexp {\S+} $line]} {
		    
		    incr i
		    
		    lappend program_to_run_list $line
		    
		    dict set program_to_run_dict $i caller [lindex $line 0] \
			command [lindex $line 1] args [lrange $line 2 end]
		    
		    #puts "caller [lindex $line 0] command [lindex $line 1] args [lrange $line 2 end]\n"
		}
		
	    }
	    
	    #dict set experiment_controller_attributes program_to_run_list $program_to_run_list
	    
	    #Store the program in the variable exp_controller_attributes
	    dict set exp_controller_attributes program_to_run_dict $program_to_run_dict
	    
	    #puts "[info class methods experimentController]\n"
	    
	    #puts "$program_to_run_list\n"
	    
	    #puts "$program_to_run_dict\n"
	    
	    #puts "The instances of all instrument SLMs:\n"
	    
	    #puts "[info class instances instrumentModule]\n"
	    
	    #puts "[[self] getSelectedInstSLM]\n"
	    
	    #puts "The instances of all equipment SLMs:\n"
	    
	    #puts "[info class instances equipmentModule]\n"
	    
	    #puts "[[self] getSelectedEquipSLM]\n"
	    
	    foreach id_line [dict keys $program_to_run_dict] {
		
		set caller [dict keys [dict get $program_to_run_dict $id_line caller]]
		
		#puts [dict get [dict get $program_to_run_dict $id caller] $caller]
		
		#puts "Line $id_line"
		
		set line [dict get $program_to_run_dict $id_line caller]
		
		set command [dict keys [dict get [dict get $program_to_run_dict $id_line caller] $caller command ] ]
		
		#puts "Caller $caller command $command\n"
		
		set return_check [[self] checkCommand "$line"]
		
		set status_command [lindex [split $return_check ";"] 0]
		
		if { $status_command == "NACK" } { set status_program NACK }

		#To mark with red color if there is NACK in return_check
		
		if { ![ string match *NACK* $return_check ] }  {
		    
		    $log_window insert end "\n$id_line $line\n"
		    $log_window insert end "$id_line - $return_check\n"
		    
		} else {

		    $log_window insert end "\n$id_line $line\n" alert_message
		    $log_window insert end "$id_line - $return_check\n" alert_message

		}
		
	    }
	    
	    
	    if { $status_program == "NACK" } {
		
		set return_event [[self] applyEventTSC eventERROR_VALIDATION]

		$log_window insert end "\nERROR VALIDATION - $return_event\n" alert_message

		$log_window see end
		
	    } else {
		
		set return_event [[self] applyEventTSC eventPROGRAM_VALIDATED]
		
		#puts "Return of validation: $return_event"

		#Tip to make the text read only
		#https://wiki.tcl-lang.org/page/Read-only+text+widget
		$program_window configure -state disabled

		$log_window insert end "PROGRAM VALIDATED - Program window disabled\n"

		$log_window see end
	    }
	    
	} else {
	    
	    #puts "ERROR: $return_event"
	    
	    set return_event [[self] applyEventTSC eventERROR_VALIDATION]

	    [self] updateStatusExperiment
	    
	    #puts "Return of event eventERROR_VALIDATION $return_event"   

	    $log_window insert end "\nERROR VALIDATION - $return_event\n" alert_message
	    
	    $log_window see end
	}

		
    }
    
    method checkCommand { line } {

	#puts "Method checkCommand of [self] for line $line\n"

	set caller [dict keys $line]

	set command [dict keys [dict get $line $caller command]]

	set args_input_user [dict get $line $caller command $command args]

	set size_args_input_user [llength $args_input_user]
	
	#puts "Caller: $caller"

	#puts "Command: $command"

	#puts "Args: $args_input_user - size: $size_args_input_user"

	
	
	if {$caller == "tsc"} {

	    #puts "The caller $caller is the TSC"

	    foreach met [info object methods [self] -all] {

		if { $command == $met } {

		    #The methods to be used in program is defined by the command
		    #oo::objdefine
		    
		    set def_command [info object definition [self] $command]

		    set args_command [lindex $def_command 0]

		    set size_args_command [llength $args_command]

		    return "ACK;$command is a command of TSC with the args: $args_command ($size_args_command)."

		}
		
	    }

	    return "NACK;Command $command is NOT a method of TSC"
	}
	
	#Check if the caller is an instrument or equipment

	set inst_server [[self] getServer inst]

	#puts "Instruments\n"

	#In case of no instrument
	if {[$inst_server getSelectedObjSLM] != "0"} {
	    
	    foreach inst_slm [$inst_server getSelectedObjSLM] {
		
		#puts "******inst_slm: $inst_slm id: [$inst_slm getID]"
		
		
		if {[$inst_server getIDModule $inst_slm] == $caller} {
		    
		    #puts "The caller $caller is an instrument\n"
		    
		    foreach cmd [$inst_server getLoadedCommands $inst_slm]  {
			
			if { $command == $cmd } {
			    
			    set list_args_cmd \
				[$inst_server getArgsCommand $inst_slm $command]
			    
			    set size_list_args_cmd [llength $list_args_cmd]
			    
			    if {$size_args_input_user == $size_list_args_cmd} {
				
				return "ACK;$command is a command of the instrument module [$inst_slm getID] with the correct number of args: $list_args_cmd ($size_list_args_cmd)"
				
			    } else {
				
				return "NACK;$command is a command of the instrument module [$inst_slm getID] BUT with the WRONG number of args: $args_input_user ($size_args_input_user). The correct is $size_list_args_cmd - $list_args_cmd"
				
			    }
			}
			
		    }
		    
		    return "NACK;Command $command is NOT a command of the instrument module [$inst_slm getID]"
		    
		}
		
	    }

	}

	#puts ""
	
	set equip_server [[self] getServer equip]

	puts "Equipments\n"

	#In case of no equipment
	if { [$equip_server getSelectedObjSLM] != "0" } {
	    
	    foreach equip_slm [$equip_server getSelectedObjSLM] {
		
		puts "equip_slm: $equip_slm id: [$equip_slm getID]"
		
		if {[$equip_server getIDModule $equip_slm] == $caller} {
		    
		    puts "The caller $caller is an equipament\n"
		    
		    foreach cmd [$equip_server getLoadedCommands $equip_slm]  {
			
			if { $command == $cmd } {
			       
			   puts "command:$command cmd:$cmd"    
			       
			       set list_args_cmd \
				   [$equip_server getArgsCommand $equip_slm $command]
			       set size_list_args_cmd [llength $list_args_cmd]
			       
			       if {$size_args_input_user == $size_list_args_cmd} {
				      
				      return "ACK;$command is a command of the equipment [$equip_slm getID] with the correct number of args: $list_args_cmd ($size_list_args_cmd)"
				      
				  } else {
				      
				      return "NACK;$command is a command of the equipment [$equip_slm getID] BUT with the WRONG number of args: $args_input_user ($size_args_input_user). The correct is $size_list_args_cmd - $list_args_cmd"
				      
				  }
				  
			   }
			   
		    }
		    
		    return "NACK;Command $command is NOT a command of the equipment module [$equip_slm getID]"	
		    
		}
	    }

	}
	
	return "NACK;The module $caller was NOT selected!"
	
    }

    method startExperiment { log_window } {

	set return_event [[self] applyEventTSC eventRUN_EXPERIMENT]

	$log_window insert end "\nReturn of event eventRUN_EXPERIMENT $return_event\n"

	$log_window see end

	if { $return_event  == "ACK" } {
	    
	    set program_validated [dict get $exp_controller_attributes program_to_run_dict]
	    
	    $log_window insert end "\nRUNNING PROGRAM\n"
	    
	    #puts "exp_controller_attributes: $exp_controller_attributes\n"
	    
	    #puts "keys or exp_controller_attributes: [dict keys $exp_controller_attributes]\n"
	    
	    #puts "Program validated:\n $program_validated\n"
	    
	    
	    foreach id [dict keys $program_validated] {

		puts "\n#Method startExperiment id:$id"

		puts "[dict get $program_validated $id]"

		#Create the transaction_id for this line of program
		set transaction_id [ clock milliseconds ]

		#Save the transaction_id in the variable program_validated
		dict set program_validated $id transaction_id $transaction_id

		puts "[dict get $program_validated $id]\n"
		
		set caller [dict keys [dict get $program_validated $id caller]]
		
		if { $caller in [ [self] getSelectedInstSLM ] } {
		    set caller_type "inst"
		    
		    set server [[self] getServer inst]
		    
		    set command [dict keys [dict get $program_validated $id caller $caller command]]
		    
		    set list_args [dict get $program_validated $id caller $caller command $command args]
		    
		    #puts [dict get $program_validated $id caller]
		    #puts "Caller of command $id: $caller is a $caller_type SLM\n"
		    #puts "Command: [dict keys [dict get $program_validated $id caller $caller command]]\n"
		    #puts "Args: $list_args"
		    
		    #set return_command [[self] execCommand $server $caller $command $list_args]

		    set return_command [[self] execCommandExp $server $caller $command $list_args $transaction_id]

		    $log_window insert end "\n$caller $command -> $return_command\n"
		    
		} elseif { $caller in [ [self] getSelectedEquipSLM ] } {
		    
		    set caller_type "equip"
		    
		    set server [[self] getServer equip]
		    
		    set command [dict keys [dict get $program_validated $id caller $caller command]]
		    
		    set list_args [dict get $program_validated $id caller $caller command $command args]
		    
		    #puts [dict get $program_validated $id caller]
		    #puts "Caller of command $id: $caller is a $caller_type SLM\n"
		    #puts "Command: [dict keys [dict get $program_validated $id caller $caller command]]\n"
		    #set return_command [[self] execCommand $server $caller $command $list_args]

		    set return_command [[self] execCommandExp $server $caller $command $list_args $transaction_id]

		    $log_window insert end "\n$caller $command -> $return_command\n"

		    $log_window see end
		    
		} elseif { $caller == "tsc" } {
		    
		    set caller_type "controller"
		    
		    set command [dict keys [dict get $program_validated $id caller $caller command]]
		    
		    set list_args [dict get $program_validated $id caller $caller command $command args]
		    
		    puts [dict get $program_validated $id caller]
		    puts "Caller of command $id: $caller is the controller of experiment\n"
		    puts "Command: [dict keys [dict get $program_validated $id caller $caller command]]\n"
		    
		    #set return_command [[self] execCommand [self] $caller $command $list_args]

		    set return_command [[self] execCommandExp [self] $caller $command $list_args $transaction_id]

		    $log_window insert end "\nTSC $command -> $return_command\n"
		    
		    $log_window see end
		}
		
	    }

	    set return_event [[self] applyEventTSC eventEND_EXPERIMENT]
	    
	    $log_window insert end "\nReturn of event eventEND_EXPERIMENT $return_event\n"
	    
	    $log_window see end
	    
	} else {
	    
	    #puts "ERROR of event eventRUN_EXPERIMENT - $return_event"
	    
	    #set return_event [[self] applyEventTSC eventSTOP_EXPERIMENT]

	    #puts "Return of event eventSTOP_EXPERIMENT $return_event"

	}
    }

    #Method to call a command in any state of SLM
    method execCommand { server caller cmd list_args } {

	#puts "[self] - method execCommand with for server:$server caller:$caller command:$cmd args $list_args\n"

	if { $caller != "tsc" } {
	    
	    #puts "The object representing $caller is [$server getObjModule $caller]\n"

	    return [$server execCommand [$server getObjModule $caller] $cmd $list_args]

	    #return [$server applyEventSLM $caller run_op "$cmd list_args"]

	} else {

	    #puts "TSC exec the command $cmd with args:$list_args"
		
	}
    }

    #method to call a command in NORMAL_OPERATION state
    method execCommandExp { server caller cmd list_args transaction_id } {

	puts "\n[self] - method execCommandExp with for server:$server caller:$caller command:$cmd args $list_args transaction_id: $transaction_id\n"

	if { $caller != "tsc" } {
	    
	    return [$server execCommandExp $caller $cmd $list_args $transaction_id ]

	    #return [$server applyEventSLM $caller run_op "$cmd list_args"]

	} else {

	       puts "TSC exec the command $cmd with args:$list_args and transaction_id: $transaction_id"

	       return [[self] $cmd {*}$list_args]
	       
		
	}
    }

    method applyEventSLM { server caller event list_args} {

	#puts "Method applyEventSLM of [info object class [self]] server - $server caller - $caller event - $event and list_args - $list_args\n "
	
	return [$server applyEventSLM $caller $event $list_args]
	
    }
    
    method stopExperiment { program_window log_window } {

	set return_event [[self] applyEventTSC eventSTOP_EXPERIMENT]
	
	#puts "Return of event eventSTOP_EXPERIMENT $return_event"

	$log_window insert end "\nReturn of event eventSTOP_EXPERIMENT $return_event\n"

	#Tip to make the text read only
	#https://wiki.tcl-lang.org/page/Read-only+text+widget
	$program_window configure -state normal

	$log_window insert end "Stoped Experiment - Program window activated.\n"

	$log_window see end
	
    }
    
    method pauseExperiment { log_window } {

	set return_event [[self] applyEventTSC eventPAUSE_EXPERIMENT]
	
	#puts "Return of event eventPAUSE_EXPERIMENT $return_event"

	$log_window insert end "\nReturn of event eventPAUSE_EXPERIMENT $return_event\n"

	$log_window see end
	
    }

    method resumeExperiment { log_window } {

	set return_event [[self] applyEventTSC eventRESUME_EXPERIMENT]
	
	#puts "Return of event eventRESUME_EXPERIMENT $return_event"

	$log_window insert end "\nReturn of event eventRESUME_EXPERIMENT $return_event\n"

	$log_window see end
	
    }

    

    method changeProgram { program_window log_window} {

	set return_event [[self] applyEventTSC eventCHANGE_PROGRAM]

	#puts "Event changeProgram - $return_event"

	#Tip to make the text read only
	#https://wiki.tcl-lang.org/page/Read-only+text+widget
	$program_window configure -state normal

	$log_window insert end "Change program - Program window activated.\n"

	$log_window see end

    }

    method updateStatusExperiment { } {

	if { [ dict exists $exp_controller_attributes obj_state] } {
	    
	    set obj_state [dict get $exp_controller_attributes obj_state]

	    #To remove the part ::state
	    set status_experiment [string range [info object class $obj_state] 7 end]
	    
	    set ::status_exp "Estado do experimento - $status_experiment"
	    
	}
	
    }
}

    
###############################################################
#Info about Machine State Object Oriented
###############################################################

#http://www.conradbock.org/statemachine.html
#Object-oriented States
#This technique starts by treating states as classes that are instantiated
#and deleted at runtime as the state machine executes,

#https://www.tclcommunityassociation.org/wub/proceedings/Proceedings-2012/SeanWoods/Lifecycle-of-Objects.pdf
#Lifecycle Object  Generators (LOG)

###############################################################
#State UNCONFIGURED
###############################################################

oo::class create stateUNCONFIGURED {

    variable tsc_state_attributes
    
    constructor { name } {

	dict set tsc_state_attributes name_object $name
    }

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method exitAction {} {

	#puts "Object [self] leaving the current state [info object class [self]]"
    }

    method eventCONFIG {} {

	#puts "Method eventCONFIG of [self] at current state [info object class [self]]"

	[self] exitAction

	oo::objdefine [self] class stateUSER_SELECT_SLM

	[self] entryAction
	
    }

    method connectInstServer {} {}

    method connectEquipServer {} {}

    method setObjState {} {}

}


###############################################################
#State USER_SELECT_SLM
###############################################################

oo::class create stateUSER_SELECT_SLM {

    variable tsc_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {

	#puts "Object [self] entering the current state [info object class [self]]"
    }

    method exitAction {} {

	#puts "Object [self] leaving the current state [info object class [self]]"
    }

    method eventCONNECT_SLM {} {

	#puts "Method eventCONNECT_SLM of [self] at current state [info object class [self]]"

	[self] exitAction
	
	oo::objdefine [self] class stateCONNECTING_TO_SLM
	
	[self] entryAction
	
    }

    method setAvailableSLM {} {}

    method setSelectedInstSLM {} {}

    method setSelectedEquipSLM {} {}
    
}

###############################################################
#State CONNECTING_TO_SLM
###############################################################

oo::class create stateCONNECTING_TO_SLM {

    variable tsc_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {

#puts "Object [self] entering the current state [info object class [self]]"
    }

    method exitAction {} {

	#puts "Object [self] leaving the current state [info object class [self]]"
    }

    method eventCONNECTION_ACCEPTED {} {

#puts "Method eventCONNECTION_ACCEPTED of [self] at current state [info object class [self]]"

	[self] exitAction

	oo::objdefine [self] class stateCONNECTED_TO_SLM

	[self] entryAction

	
    }

    method eventERROR_CONNECTION {} {

#puts "Method eventERROR_CONNECTION of [self] at current state [info object class [self]]"

	[self] exitAction

	oo::objdefine [self] class stateUSER_SELECT_SLM

	[self] entryAction

	
    }

}


###############################################################
#State CONNECTED_TO_SLM
###############################################################

oo::class create stateCONNECTED_TO_SLM {

    variable tsc_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method eventINIT_SLM {} {

#puts "Method eventINIT_SLM of [self] at current state [info object class [self]]"

	[self] exitAction

	oo::objdefine [self] class stateINITING_SLM

	[self] entryAction

    }
    
    method entryAction {} {

#puts "Object [self] entering the current state [info object class [self]]"
    }

    method exitAction {} {

#puts "Object [self] leaving the current state [info object class [self]]"
    }

}

###############################################################
#State INITING_SLM
###############################################################

oo::class create stateINITING_SLM {

    variable tsc_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method eventERROR_INIT {} {

#puts "Method errorINIT_SLM of [self] at current state [info object class [self]]"

	[self] exitAction

	oo::objdefine [self] class stateCONNECTED_TO_SLM

	[self] entryAction

    }

    method eventEND_INIT {} {

#puts "Method errorEND_INIT of [self] at current state [info object class [self]]"

	[self] exitAction

	oo::objdefine [self] class stateCONFIGURED

	[self] entryAction

	[self] exitAction

	oo::objdefine [self] class stateUSER_PROGRAMMING

	[self] entryAction

    }
    
    method entryAction {} {

#puts "Object [self] entering the current state [info object class [self]]"

	
	
    }

    method exitAction {} {

#puts "Object [self] leaving the current state [info object class [self]]"
    }

}

###############################################################
#State CONFIGURED
###############################################################

oo::class create stateCONFIGURED {

    variable tsc_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method entryAction {} {

#puts "Object [self] entering the current state [info object class [self]]"
    }

    method exitAction {} {

#puts "Object [self] leaving the current state [info object class [self]]"
    }

}

###############################################################
#State USER_PROGRAMMING
###############################################################

oo::class create stateUSER_PROGRAMMING {

    superclass stateCONFIGURED
    
    variable tsc_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {

#puts "Object [self] entering the current state [info object class [self]]"
    }

    method exitAction {} {

#puts "Object [self] leaving the current state [info object class [self]]"
    }


    method eventVALIDATE_PROGRAM {} {
	
#puts "Method eventValidate_PROGRAM of [self] at current state [info object class [self]]"
	
	[self] exitAction
	
	oo::objdefine [self] class stateVALIDATING
	
	[self] entryAction
	
    }

    
}

###############################################################
#State VALIDATING
###############################################################

oo::class create stateVALIDATING {

    superclass stateCONFIGURED
    
    variable tsc_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {

#puts "Object [self] entering the current state [info object class [self]]"
    }

    method exitAction {} {

#puts "Object [self] leaving the current state [info object class [self]]"
    }


    method eventPROGRAM_VALIDATED {} {
	
#puts "Method eventPROGRAM_VALIDATED of [self] at current state [info object class [self]]"
	
	[self] exitAction
	
	oo::objdefine [self] class stateREADY_TO_RUN_EXPERIMENT
	
	[self] entryAction
	
    }

    method eventERROR_VALIDATION {} {
	
#puts "Method eventERROR_VALIDATION of [self] at current state [info object class [self]]"
	
	[self] exitAction
	
	oo::objdefine [self] class stateUSER_PROGRAMMING
	
	[self] entryAction
	
    }

}

###############################################################
#State READY_TO_RUN_EXPERIMENT
###############################################################

oo::class create stateREADY_TO_RUN_EXPERIMENT {

    superclass stateCONFIGURED
    
    variable tsc_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {

#puts "Object [self] entering the current state [info object class [self]]"
    }

    method exitAction {} {

#puts "Object [self] leaving the current state [info object class [self]]"
    }


    method eventRUN_EXPERIMENT {} {
	
#puts "Method eventRUN_EXPERIMENT of [self] at current state [info object class [self]]"
	
	[self] exitAction
	
	oo::objdefine [self] class stateRUNNING_EXPERIMENT
	
	[self] entryAction
	
    }

    method eventCHANGE_PROGRAM {} {
	
#puts "Method eventCHANGE_PROGRAM of [self] at current state [info object class [self]]"
	
	[self] exitAction
	
	oo::objdefine [self] class stateUSER_PROGRAMMING
	
	[self] entryAction
	
    }

}

###############################################################
#State RUNNING_EXPERIMENT
###############################################################

oo::class create stateRUNNING_EXPERIMENT {

    superclass stateCONFIGURED
    
    variable tsc_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {

#puts "Object [self] entering the current state [info object class [self]]"
    }

    method exitAction {} {

#puts "Object [self] leaving the current state [info object class [self]]"
    }


    method eventEND_EXPERIMENT {} {
	
#puts "Method eventEND_EXPERIMENT of [self] at current state [info object class [self]]"
	
	[self] exitAction
	
	oo::objdefine [self] class stateREADY_TO_RUN_EXPERIMENT
	
	[self] entryAction
	
    }

    method eventSTOP_EXPERIMENT {} {
	
#puts "Method eventSTOP_EXPERIMENT of [self] at current state [info object class [self]]"
	
	[self] exitAction
	
	oo::objdefine [self] class stateUSER_PROGRAMMING
	
	[self] entryAction
	
    }

    method eventPAUSE_EXPERIMENT {} {
	
#puts "Method eventPAUSE_EXPERIMENT of [self] at current state [info object class [self]]"
	
	[self] exitAction
	
	oo::objdefine [self] class statePAUSED_EXPERIMENT
	
	[self] entryAction
	
    }

}

###############################################################
#State PAUSED_EXPERIMENT
###############################################################

oo::class create statePAUSED_EXPERIMENT {

    superclass stateCONFIGURED
    
    variable tsc_state_attributes

    method applyStateEvent { event } {

	if { [catch { [self] $event } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }

    method checkCommandTSC { cmd } {

	if { [catch { [self] $cmd } result] != 0 } {
	    
	    return "NACK;$result"

	} else {

	    return "ACK"

	}
	
    }
    
    method entryAction {} {

#puts "Object [self] entering the current state [info object class [self]]"
    }

    method exitAction {} {

#puts "Object [self] leaving the current state [info object class [self]]"
    }


    method eventRESUME_EXPERIMENT {} {
	
#puts "Method eventEND_EXPERIMENT of [self] at current state [info object class [self]]"
	
	[self] exitAction
	
	oo::objdefine [self] class stateRUNNING_EXPERIMENT
	
	[self] entryAction
	
    }

}

###############################################################
#Graphical User Interface
###############################################################

wm title . "Sistema para Monitoramento e Controle de Experimentos"

#To prevent resizing of windows

wm resizable . 0 0

global big_font medium_font 

set big_font {Times 22}

set medium_font {Times 16}

###################################################################
#Window to select and set the equipments and instruments to be used
###################################################################

proc selectEquipInst { } {

    global big_font medium_font f_sel_equip_inst experiment_controller

    set f_sel_equip_inst [frame .f_sel_equip_inst]

    set message_top [label $f_sel_equip_inst.msg_top -font $big_font -text "Selecione os Equipamentos e Instrumentos que serão usados"]

    #puts [$experiment_controller getAvailableSLM]

    foreach slm [$experiment_controller getAvailableSLM] {

	#puts "SLM: $slm"

    }

    #puts "Available instrument SLM: [$experiment_controller getAvailableInstSLM]"
    
    #puts "Available equipment SLM: [$experiment_controller getAvailableEquipSLM]"
    
    set f_sel_equip [labelframe $f_sel_equip_inst.set_equip -font $big_font -text "Equipamentos"]

    foreach {slm_id slm_name} [$experiment_controller getAvailableEquipSLM] {

	#puts "Loop foreach for equip: slm_id:$slm_id and name:[lindex $slm_name 1]"

	set check_$slm_id \
	    [checkbutton $f_sel_equip.chk_$slm_id -text "[lindex $slm_name 1]" \
		 -font $medium_font -relief flat \
		 -variable use_$slm_id \
		 -onvalue 1 \
		 -offvalue 0 \
		]
	grid [set check_$slm_id]
    }

    
    
    set f_sel_inst [labelframe $f_sel_equip_inst.set_inst -font $big_font -text "Instrumentos"]

    foreach {slm_id slm_name} [$experiment_controller getAvailableInstSLM] {

	#puts "Loop foreach for inst: slm_id:$slm_id and name:[lindex $slm_name 1]"

	set check_$slm_id \
	    [checkbutton $f_sel_inst.chk_$slm_id -text "[lindex $slm_name 1]" \
		 -font $medium_font -relief flat \
		 -variable use_$slm_id \
		 -onvalue 1 \
		 -offvalue 0 \
		]
	grid [set check_$slm_id]
    }


    set button_cont [button $f_sel_equip_inst.btn_cont -text "Continuar" -font $medium_font -command setEquipInst]

    set button_exit [button $f_sel_equip_inst.btn_exit -text "  Sair  " -font $medium_font -command exit]
    
    grid $message_top -columnspan 2

    grid $f_sel_equip $f_sel_inst

    grid $button_exit $button_cont -pady 10
    
    grid $f_sel_equip_inst

    ##########################################
    #Change the state of controller
    #from UNCONFIGURED to USER_SELECT_SLMs
    ##########################################

    $experiment_controller applyEventTSC eventCONFIG

}



proc setEquipInst {} {
    
    global f_sel_equip_inst experiment_controller

    destroy $f_sel_equip_inst
    
    
    ##########################################
    #Change the state of controller
    #from USER_SELECT_SLMs to CONNECTING_TO_SLM
    ##########################################

    $experiment_controller applyEventTSC eventCONNECT_SLM

    set error_connection 0
    
    #Set the equipment modules SLMs

    
    set equip_server [$experiment_controller getServer equip]
    
    foreach {slm_id slm_name} [$experiment_controller getAvailableEquipSLM] {
	
	#puts "Loop foreach in setEquipInst for equip: slm_id:$slm_id and name:[lindex $slm_name 1]"

	if { [set ::use_$slm_id] } {

	    #puts "Use SLM [lindex $slm_name 1]"

	    $experiment_controller setSelectedEquipSLM $slm_id

	    #In this command the instrumentServer create the object for inst SLM
	    #create the corresponding state, aggregate the obj state to obj module

	    set return_start_slm_connection [$equip_server startConnection $slm_id]

	    #Implement test if using return
	    
	   #puts "\n----return_start_slm_connection: $return_start_slm_connection\n"
	    if { $return_start_slm_connection == "ACK" } {

		$experiment_controller setActiveEquipSLM $slm_id

	    } else {

		set error_connection 1

	    }
	}
    }

    

    #Set the instrument modules SLMs
    
    set inst_server [$experiment_controller getServer inst]

    foreach {slm_id slm_name} [$experiment_controller getAvailableInstSLM] {
	
	#puts "Loop foreach in setEquipInst for inst: slm_id:$slm_id and name:[lindex $slm_name 1]"

	if { [set ::use_$slm_id] } {

	    $experiment_controller setSelectedInstSLM $slm_id

	    #puts "Use SLM [lindex $slm_name 1]"

	    #In this command the instrumentServer create the object for inst SLM
	    #create the corresponding state, aggregate the obj state to obj module

	    set return_start_slm_connection [$inst_server startConnection $slm_id]

	   #puts "\n----return_start_slm_connection: $return_start_slm_connection\n"

	    if { $return_start_slm_connection == "ACK" } {

		$experiment_controller setActiveInstSLM $slm_id

	    } else {

		set error_connection 1

	    }
	    
	    
	}

    }


    

    ##########################################################
    #Change the state of controller
    #from CONNECTING_TO_SLM to CONNECTED_TO_SLM
    #but check any error before
    ##########################################################

    if { $error_connection } {

	$experiment_controller applyEventTSC eventERROR_CONNECTION

    } else {

	#Inicialize the variable to register any error
	set error_init_setup {}

	$experiment_controller applyEventTSC eventCONNECTION_ACCEPTED

	
	if { [$experiment_controller getActiveEquipSLM] != "0" } {
	    
	    foreach active_slm [$experiment_controller getActiveEquipSLM] {
		
		#INIT event for the equipments
		set return_event [$experiment_controller applyEventSLM $equip_server $active_slm init NO_ARG]
		
	#puts "\n      *return_event - $return_event of event init to equip_server\n"
		
		if { $return_event == "ACK" } {
		    
		    $experiment_controller applyEventSLM $equip_server $active_slm end_init NO_ARG
		    
		    #SETUP event for equipments
		    set return_event [$experiment_controller applyEventSLM $equip_server $active_slm setup NO_ARG]
		    
		   #puts "\n      **return_event - $return_event of event setup to equip_server\n"
		    
		    if { $return_event == "ACK" } {
			
			$experiment_controller applyEventSLM $equip_server $active_slm end_setup NO_ARG
			
		    } else {
			
			$experiment_controller applyEventSLM $equip_server $active_slm error_setup NO_ARG

			lappend error_init_setup "$active_slm error_setup"
			
		    }
		    
		} else {
		    
		    $experiment_controller applyEventSLM $equip_server $active_slm error_init NO_ARG

		    lappend error_init_setup "$active_slm error_init"
		}
		
	    }
	    
	}


	if { [$experiment_controller getActiveInstSLM] != "0" } {
	    
	    foreach active_slm [$experiment_controller getActiveInstSLM] {
		
		#INIT event for instruments
		set return_event [$experiment_controller applyEventSLM $inst_server $active_slm init NO_ARG]
		
	#puts "\n      *return_event - $return_event of event init to inst_server\n"
		
		if { $return_event == "ACK" } {
		    
		    $experiment_controller applyEventSLM $inst_server $active_slm end_init NO_ARG
		    
		    #SETUP event for instruments
		    set return_event [$experiment_controller applyEventSLM $inst_server $active_slm setup NO_ARG]
		    
		   #puts "\n      **return_event - $return_event of event setup to inst_server\n"
		    
		    if { $return_event == "ACK" } {
			
			$experiment_controller applyEventSLM $inst_server $active_slm end_setup NO_ARG


			
		    } else {
			
			$experiment_controller applyEventSLM $inst_server $active_slm error_setup NO_ARG

			lappend error_init_setup "$active_slm error_setup" 
			
		    }
		    
		} else {
		    
		    $experiment_controller applyEventSLM $inst_server $active_slm error_init NO_ARG

		    lappend error_init_setup "$active_slm error_setup" 
		}
		
	    }

	}

	##########################################################
	#Change the state of controller
	#from CONNECTED_TO_SLM to INITING_SLM
	##########################################################

	if { $error_init_setup == "" } {
	    
	    #Check the state permission
	    
	    $experiment_controller applyEventTSC eventINIT_SLM

	    ##########################################################
	    #Change the state of controller
	    #from INITING_SLM to CONFIGURED
	    ##########################################################
	    $experiment_controller applyEventTSC eventEND_INIT	

	}
    
	
    }

    #Procedure to program and run the experiment
    programExperiment

}

proc programExperiment {} {

    global big_font medium_font experiment_controller available_resources

    #puts "\n#Procedure programExperiment#\n"
    
    set f_config_exp [frame .config_experiment]

    set message_top [label $f_config_exp.msg_top -font $big_font -text "Programe o experimento com os comandos disponíveis, verifique a sintaxe e inicie a execução."] 

    set message_status_exp [label $f_config_exp.msg_status_exp \
				-font $big_font -textvariable ::status_exp]

    set ::status_exp "Estado do experimento:"
    
   #puts " ***Active instrument SLMs: [$experiment_controller getActiveInstSLM]\n"

   #puts " ***Active equipment SLMs: [$experiment_controller getActiveEquipSLM]\n"

    #Frame to show the availble commands from selected SLMs and from the TSC
    set f_commands_available [labelframe $f_config_exp.commands]

    set text_list_commands [text $f_commands_available.list_commands \
				-height 10 -width 50 \
				-undo 1 -autosep 1 -spacing1 10]

    set scroll_list_commands [scrollbar $f_commands_available.scroll_list_commands \
				    -command "$text_list_commands yview"]

    $text_list_commands configure -yscrollcommand "$scroll_list_commands set"

    #Tip - http://pages.cpsc.ucalgary.ca/~saul/personal/archives/Tcl-Tk_stuff/tcl_examples/
    #Create a tag to color alert messages
    $text_list_commands tag configure alert_message -foreground red
    
    #Menubutton of available commands
    
    set menu_file_available_commands [menubutton $f_commands_available.file -font $medium_font \
			       -text "Módulos e Comandos Disponíveis" -relief raised \
			       -menu $f_commands_available.file.options]

    menu $f_commands_available.file.options

    $f_commands_available.file.options add command -font $medium_font -label "Gravar" \
	-command "save_file $text_list_commands"
    
    #########################################################################
    #Set the available commands of selected equipments and instruments (SLMs)
    #########################################################################

    $experiment_controller setAllAvailableCommands $available_resources

    #Widget text to include the list of commands $text_list_commands

    #Call the method getAllAvailableCommands using inst_commands or
    #equip_commands as argument
    $text_list_commands insert end "############################\n"
    $text_list_commands insert end "Comandos para os instrumentos\n"
    $text_list_commands insert end "############################\n"

    #The method getAvailableInstSLM returns all inst SLM independent
    #of user selection
    set list_available_inst [$experiment_controller getAvailableInstSLM]

    #The method getAllAvailableCommands returns only the commands of
    #selected SLMs
    set inst_commands \
	[$experiment_controller getAllAvailableCommands inst_commands]

   #puts " **Procedure programExperiment list_available_inst - $list_available_inst and inst_commands - $inst_commands\n"
    
    foreach slm [dict keys $inst_commands] {

	set i [lsearch -exact $list_available_inst $slm]

	$text_list_commands insert end "############################\n"
	$text_list_commands insert end \
	    "$slm ([lindex $list_available_inst $i+1])\n"

	#To alert any error in conection
	set list_active_inst [$experiment_controller getActiveInstSLM]
	
	if { $slm ni $list_active_inst } {
	    
	    $text_list_commands insert end "ERROR - error conection with $slm\n" alert_message
	}
	
	foreach command [dict keys [dict get $inst_commands $slm]] {
	    
	    #$text_list_commands insert end \
		#"$command ([dict get $inst_commands $slm $command])\n"
	        
	    foreach arg [dict keys [dict get $inst_commands $slm $command]] {

		$text_list_commands insert end \
		    "$command ([dict get $inst_commands $slm $command $arg])\n"
		
	    }
	    
	}
	
    }

    $text_list_commands insert end "#############################\n"
    $text_list_commands insert end "Comandos para os equipamentos\n"
    $text_list_commands insert end "#############################\n"
    
    set list_available_equip [$experiment_controller getAvailableEquipSLM]

    set equip_commands \
	[$experiment_controller getAllAvailableCommands equip_commands]

    foreach slm [dict keys $equip_commands] {

	set i [lsearch -exact $list_available_equip $slm]

	$text_list_commands insert end "############################\n"
	$text_list_commands insert end \
	    "$slm ([lindex $list_available_equip $i+1])\n"

	#To alert any error in conection
	set list_active_equip [$experiment_controller getActiveEquipSLM]

		if { $slm ni $list_active_equip } {

		    $text_list_commands insert end "ERROR - error conection with $slm\n" alert_message
		}
	
	foreach command [dict keys [dict get $equip_commands $slm]] {
	    
	    #$text_list_commands insert end \
	    #	"$command ([dict get $equip_commands $slm $command])\n"

	    foreach arg [dict keys [dict get $equip_commands $slm $command]] {

		$text_list_commands insert end \
		    "$command ([dict get $equip_commands $slm $command $arg])\n"
		
	    }
	    
	}
	
    }

    ###############################################################
    #Include the commands of TSC
    ###############################################################

    foreach caller [dict keys $available_resources] {
	
	if { $caller == "tsc" } {

	    $text_list_commands insert end "############################\n"
	    set name [dict get $available_resources $caller name]
	    
	    $text_list_commands insert end "$caller (name $name)\n"
	    
	    #set more_info [dict get $available_resources $caller]
	    
	    #$text_list_commands insert end "Info: $more_info\n"
	    
	    set commands_info [dict get $available_resources $caller command]
	    
	    #$text_list_commands insert end "Comando info: $commands_info\n"
	    
	    foreach cmd [dict keys $commands_info] {

		$text_list_commands insert end "################\n"
		$text_list_commands insert end "Comando: $cmd\n"
		
		#set args [dict get $commands_info $cmd args]
		
		#$text_list_commands insert end "Parâmetros: $args\n"

		foreach arg [dict keys [dict get $commands_info $cmd args]] {

		    $text_list_commands insert end \
			"Parâmetro: $arg [dict get $commands_info $cmd args $arg]\n"

		}
		
	    }
	    
	}
	
    }
    
    
    #Tip to make the text read only
    #https://wiki.tcl-lang.org/page/Read-only+text+widget
    $text_list_commands configure -state disabled
    
    #Frame to program the experiment
    set f_program_exp [labelframe $f_config_exp.program ]

    set text_program_exp [text $f_program_exp.program_exp \
				-height 10 -width 50 \
				-undo 1 -autosep 1 -spacing1 10]

    set scroll_program_exp [scrollbar $f_program_exp.scroll_program_exp \
				    -command "$text_program_exp yview"]

    $text_program_exp configure -yscrollcommand "$scroll_program_exp set"

        
    #Frame to log the events along the experiment
    #and the status of SLMs along the experiment
    set f_log_exp [labelframe $f_config_exp.log \
		       -text "" \
		       -font $medium_font ]

    set text_log_exp [text $f_log_exp.log_exp \
				-height 10 -width 50 \
				-undo 1 -autosep 1 -spacing1 10]

    set scroll_log_exp [scrollbar $f_log_exp.scroll_log_exp \
				    -command "$text_log_exp yview"]

    $text_log_exp configure -yscrollcommand "$scroll_log_exp set"

    
    
    #Execute the program
    set button_exec_exp [button $f_config_exp.btn_exec_exp -text "Iniciar Experimento" -font $medium_font]

    $button_exec_exp configure -command \
	[list $experiment_controller startExperiment $text_log_exp]

    set button_stop_exp [button $f_config_exp.btn_stop_exp -text "Parar Experimento" -font $medium_font]

    $button_stop_exp configure -command \
	[list $experiment_controller stopExperiment $text_program_exp $text_log_exp]

    set button_exit [button $f_config_exp.btn_exit -text "Fechar Program  " -font $medium_font -command exit]


    #Menubutton of program
    
    set menu_file_program [menubutton $f_program_exp.file -font $medium_font \
			       -text "Programa do Experimento" -relief raised \
			       -menu $f_program_exp.file.options]

    menu $f_program_exp.file.options

    $f_program_exp.file.options add command -font $medium_font -label "Abrir Programa" \
	-command "open_file $text_program_exp"

    $f_program_exp.file.options add command -font $medium_font -label "Gravar Programa" \
	-command "save_file $text_program_exp"
    
    $f_program_exp.file.options add command -font $medium_font -label "Validar Programa" \
	-command [list $experiment_controller validateProgram $text_program_exp $text_log_exp]

    $f_program_exp.file.options add command -font $medium_font -label "Iniciar Experimento" \
	-command [list $experiment_controller startExperiment $text_log_exp]

    $f_program_exp.file.options add command -font $medium_font -label "Pausar Experimento" \
	-command [list $experiment_controller pauseExperiment $text_log_exp]

    $f_program_exp.file.options add command -font $medium_font -label "Continuar Experimento" \
	-command [list $experiment_controller resumeExperiment $text_log_exp]

    $f_program_exp.file.options add command -font $medium_font -label "Parar Experimento" \
	-command [list $experiment_controller stopExperiment $text_program_exp $text_log_exp]

    $f_program_exp.file.options add command -font $medium_font -label "Alterar Programa" \
	-command [ list $experiment_controller changeProgram $text_program_exp $text_log_exp ]
    
    

    #Menubutton of log exp
    
    set menu_file_log_exp [menubutton $f_log_exp.file -font $medium_font \
			       -text "Eventos do Experimento" -relief raised \
			       -menu $f_log_exp.file.options]

    menu $f_log_exp.file.options

    $f_log_exp.file.options add command -font $medium_font -label "Gravar" \
	-command "save_file $text_log_exp"

    $f_log_exp.file.options add command -font $medium_font -label "Limpar tela" \
	-command "clean_window $text_log_exp"

    grid $menu_file_available_commands -sticky ew
    
    grid $text_list_commands $scroll_list_commands -sticky ns

    grid $menu_file_program -sticky ew
    
    grid $text_program_exp $scroll_program_exp -sticky ns
    
    grid $menu_file_log_exp -sticky ew

    grid $text_log_exp $scroll_log_exp -sticky ns
    
    grid $message_top -columnspan 3

    grid $message_status_exp -columnspan 3
    
    grid $f_commands_available $f_program_exp $f_log_exp
    
    grid $button_exec_exp $button_stop_exp $button_exit -pady 5
    
    grid $f_config_exp

    #To update the status when open the GUI
    #$experiment_controller changeProgram $text_program_exp $text_log_exp
    $experiment_controller updateStatusExperiment
}

#****************************************************************************
#Procedure to open file
#****************************************************************************

proc open_file { text_window } {
    set type_file {
	{"Arquivos de Texto"    { .txt } }
	{"Arquivos de Leituras" { .csv } }
	{"Arquivos de Eventos" { .log } }
	{"Todos os arquivos" * }
    }
    
    if {[catch {set name_file [tk_getOpenFile -filetypes $type_file -initialdir pwd]} result] !=0 } {
	return
    }
    if {[catch {set id_file [open $name_file r]} result] != 0 } {
	return
    }
    set data [read $id_file ]
    close $id_file
    $text_window delete 1.0 end
    $text_window insert end $data
}

#***********************************************************************
#Procedure to save file
#***********************************************************************

proc save_file { text_window } {
    
    set data [$text_window get 1.0 {end -1c}]
    set type_file {
	{"Arquivos de Texto"    { .txt } }
	{"Arquivos de Eventos"    { .log } }
	{"Todos os arquivos" * }
    }
    
    set name_file [tk_getSaveFile -filetypes $type_file -initialdir pwd]
    
    if {[catch {set id_file [open $name_file w]} result] != 0} {

	   return

       } else {
	   
	   puts $id_file $data
	   close $id_file
       }

}

#****************************************************************************
#Procedure to clean window
#****************************************************************************

proc clean_window { text_window } {

    $text_window delete 1.0 end

}

###############################################################
#Create the list of instruments and equipments available
###############################################################

#If any slm has a slm with no arguments, use the constant NO_ARG to avoid errors
#dict set available_resources slm_xx name "Módulo sem argumentos"
#dict set available_resources slm_xx type "instrument or equipment"
#dict set available_resources slm_xx command cmd args NO_ARG


dict set available_resources slm_00 name "Multifotômetro"
dict set available_resources slm_00 type instrument
dict set available_resources slm_01 name "Válvula 01"
dict set available_resources slm_01 type equipment
dict set available_resources slm_02 name "Válvula 02"
dict set available_resources slm_02 type equipment
dict set available_resources slm_03 name "WebCam"
dict set available_resources slm_03 type instrument
dict set available_resources tsc    name "Controlador de Tarefas"
dict set available_resources tsc    type task_controller

#Info about an integrated SLM, the syringe pump

dict set available_resources slm_04 name "Bomba de Seringa"
dict set available_resources slm_04 type equipment

#Set info about the commands of slm_00

dict set available_resources slm_00 command readABS args color {uv blue green red}

dict set available_resources slm_00 command readFLUOR args color_emitter {uv blue green red}
dict set available_resources slm_00 command readFLUOR args color_detector {uv blue green red}

dict set available_resources slm_00 command readTURB args color {uv blue green red}


#Set info about the commands of slm_03

dict set available_resources slm_03 command configWebCam args resolution
dict set available_resources slm_03 command getImage args size

dict set available_resources slm_01 command setPosition args \
    "position;user" { 1 2 3 }
dict set available_resources slm_01 command setPosition args \
    "angle;arduino" { 0 45 90 }
dict set available_resources slm_01 command setPosition args \
    "id;arduino" S1

#Set info about the commands of slm_02

dict set available_resources slm_02 command setPosition args \
    "position;user" { 1 2 3 }
dict set available_resources slm_02 command setPosition args \
    "angle;arduino" { 0 45 90 }
dict set available_resources slm_02 command setPosition args \
   "id;arduino" S1

#Set info about the commands of slm_04

#startPump;[id_pump];[set_point];[interval];[id_transaction]
dict set available_resources slm_04 command startPump args id_pump { p1 p2 }
dict set available_resources slm_04 command startPump args set_point { sp }
dict set available_resources slm_04 command startPump args interval  { int }


#Set info about the commands of TSC

#TSC startReading

#The args of command starReading are organized in a nested structure
#to reproduce the nested structure of caller - command - args
dict set available_resources tsc command startReading args \
    sub_caller {slm_00 slm_01 ...} \
    sub_command { \
		      readABS {color} \
		      readFluor { color_emitter color_detector } \
		      readTURB {color} } \
    sub_args { abs_color fluor_emitter_detector turb_color ... }

#TSC stopReading

dict set available_resources tsc command stopReading args \
    sub_caller {slm_00 slm_01 ...} \
    sub_command { \
		      readABS {color} \
		      readFluor { color_emitter color_detector } \
		      readTURB {color} } \
    sub_args { abs_color fluor_emitter_detector turb_color ... }

#TSC waitUntil 
#The args are not nested

dict set available_resources tsc command waitUntil args \
    variable_to_monitor { abs_color fluor_emitter_detector turb_color ... }
dict set available_resources tsc command waitUntil args \
    operator { > < }
dict set available_resources tsc command waitUntil args \
    value { number }

#TSC startMonitorVar

dict set available_resources tsc command startMonitorVar args \
    variable_to_monitor { abs_color fluor_emitter_detector turb_color ... }


#TSC stopMonitorVar

dict set available_resources tsc command stopMonitorVar args \
    variable_to_monitor { abs_color fluor_emitter_detector turb_color ... }

#TSC waitTime

dict set available_resources tsc command waitTime args \
    time_to_wait { t }
dict set available_resources tsc command waitTime args \
    time_unit { h m s }

#puts "#SLMs available#"
#puts $available_resources
#puts [dict keys $available_resources]

###############################################################
# Add the current folder to the search path if it
# is not already in the set of directories to be searched.
###############################################################

#puts "Path for SLMs modules before [::tcl::tm::list]\n"

if {[lsearch [::tcl::tm::path list] [pwd]] < 0} {
  ::tcl::tm::path add [pwd]
}

::tcl::tm::path add lib_slm/

#puts "Path for SLM modules after [::tcl::tm::list]\n"

###############################################################
#Create the instances of experimentController, stateUNCONFIGURED
#instrumentServer and equipmentServer
###############################################################

set experiment_controller [experimentController new]

puts "experimentController: $experiment_controller"

###############################################################
#Aggregate state of TSC to experimentController
###############################################################

set obj_state_tsc [stateUNCONFIGURED new tsc]

puts "Object state of TSC: $obj_state_tsc"

$experiment_controller setObjState $obj_state_tsc

#Create the instance of instrumentServer

set instrument_server [instrumentServer new]

puts "instrumentServer: $instrument_server"

#Create the instance of EquipmentServer

set equipment_server [equipmentServer new]

puts "equipmentServer: $equipment_server"


###############################################################
#Connect equipmentServer and instrumentServer to experimentController
###############################################################

$experiment_controller connectInstServer $instrument_server

$experiment_controller connectEquipServer $equipment_server

###############################################################
#Create the instance of dataFlowController and connect to
#instrumentServer and equipmentServer
###############################################################

set data_flow_controller [dataFlowController new]

$instrument_server setDataFlowController $data_flow_controller

puts "\nData flow controller of instrumentServer: [$instrument_server getDataFlowController]\n"

$equipment_server setDataFlowController $data_flow_controller

puts "\nData flow controller of equipmentServer: [$equipment_server getDataFlowController]\n"

###############################################################
#Set the available equipments and instruments (SLMs)
#to their respective servers
###############################################################

$equipment_server setAvailableEquip $available_resources

#puts [$equipment_server getAvailableEquip]

$instrument_server setAvailableInst $available_resources

#puts [$instrument_server getAvailableInst]

###############################################################
#Transfer the info about the available SLMs to the
#experiment controller
###############################################################

$experiment_controller setAvailableSLM [$equipment_server getAvailableEquip]

$experiment_controller setAvailableSLM [$instrument_server getAvailableInst]

###################################################################
#Create aditional commands for TSC (methods in experimentController
###################################################################

oo::objdefine $experiment_controller {

    #args concatenate the arguments that were not assigned
    #to declared variables into a list and assign that list
    #to the variable args

    method startReading { sub_caller sub_command args } {
	
	return "sub_caller: $sub_caller sub_command: $sub_command and args: $args "
	
    }
    
    method stopReading { sub_caller sub_command args } {
	
	return "$sub_caller sub_command: $sub_command and args: $args "
	
    }

    method waitUntil { var_to_monitor operator value } {

	return "var_to_monitor: $var_to_monitor operator: $operator and value: $value"
	
    }

    method startMonitorVar { var_to_monitor } {

	return "var_to_monitor: $var_to_monitor"

    }

    method stopMonitorVar { var_to_monitor } {

	return "var_to_monitor: $var_to_monitor"
    }

    #Procedimento para esperar o tempo "t" na
    #unidade "u" que pode ser {h, H, m, M, s ou S}
    #Method to wait a time defined in unit wich may be
    #(h, H, m, M, s ou S)
    method waitTime { time_to_wait time_unit } {

	puts "\nMethod waitTime with $time_to_wait and $time_unit\n"
	
	set time_unit [string toupper $time_unit]
	
	switch $time_unit {
	    "S" { set time_to_wait [expr $time_to_wait * 1000] }
	    "M" { set time_to_wait [expr $time_to_wait * 60000] }
	    "H" { set time_to_wait [expr $time_to_wait * 3600000] }
	}
	
	
	set id_after_wait_time [after $time_to_wait [list set vwait_monitor 0]]

	dict set exp_controller_attributes id_after_wait_time $id_after_wait_time

	set vwait_monitor 1

	vwait vwait_monitor

	return "$time_to_wait milliseconds"
	
    }

}

###############################################################
#Init GUI
###############################################################

selectEquipInst



